from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
import time
from typing import Any, Callable

from loguru import logger
from sprite_ai.audio.speach.speaker import Speaker
from sprite_ai.audio.transcriber import Transcriber
from sprite_ai.language.language_model import LanguageModel


@dataclass
class Assistant:
    transcriber: Transcriber
    language_model: LanguageModel
    speaker: Speaker
    on_transcription: Callable | None = None

    def foward(
        self, user_request: BytesIO | str, session_id: str = 'default'
    ) -> str:
        """Processes user request

        Args:
            user_request (BytesIO | str): Can be an audio BytesIO or an input string,
            audios will be converted to text automatically.
        """
        if not isinstance(user_request, str) and not isinstance(
            user_request, BytesIO
        ):
            raise TypeError(
                f'Invalid type "{type(user_request)}", user_request must be BytesIO or str'
            )

        if isinstance(user_request, BytesIO):
            user_request = self.transcriber(user_request)
            self.on_transcription(user_request)

        logger.info('[STARTED] Language model inference')
        awnser_started = time.time()

        awnser = self.language_model(user_request, session_id)

        awnser_elapsed = time.time() - awnser_started
        logger.info('[FINISHED] Language model inference')
        logger.info(f'[LanguageModel | Elapsed Time] {awnser_elapsed}')

        logger.info('[STARTED] Speach generation')
        speach_generation_started = time.time()

        self.speaker(awnser)

        speach_generation_elapsed = time.time() - speach_generation_started
        logger.info('[FINISHED] Speach generation')
        logger.info(
            f'[SpeachGeneration | Elapsed Time] {speach_generation_elapsed}'
        )

        return awnser

    def clear_state(self, session_id: str = 'default'):
        self.language_model.clear_memory(session_id)

    def __call__(
        self, user_request: BytesIO | str, session_id: str = 'default'
    ) -> str:
        return self.foward(user_request, session_id)



from __future__ import annotations
from pathlib import Path

import pickle
from dataclasses import dataclass
from typing import Protocol

from langchain.chains import LLMChain
import yaml


class LanguageModel:
    def __init__(self, llm_chain: LLMChain):
        self.llm_chain = llm_chain

    def foward(self, prompt: str, session_id: str) -> str:
        if self.llm_chain:
            awnser = self.llm_chain.invoke(
                {'input': prompt},
                config={'configurable': {'session_id': session_id}},
            )
        else:
            raise RuntimeError('Failed to load llm model')

        return awnser.content

    def __call__(self, prompt: str, session_id: str) -> str:
        return self.foward(prompt, session_id)

    def messages(self, session_id: str):
        return self.llm_chain.get_session_history(session_id).messages

    def clear_memory(self, session_id: str):
        self.llm_chain.get_session_history(session_id).clear()



from pathlib import Path
import yaml
from pydantic import BaseModel

from sprite_ai.language.chat_message import ChatMessage


class ChatState(BaseModel):
    chat_history: list[ChatMessage] = []

    @classmethod
    def from_file(cls, file_location: str | Path):
        file_location = Path(file_location)
        with file_location.open() as state_file:
            chat_state = yaml.safe_load(state_file)
        return cls(**chat_state)

    def to_file(self, file_location: str | Path):
        with file_location.open('w') as state_file:
            yaml.safe_dump(self.model_dump(), state_file)



from pathlib import Path
from typing import Callable
from sprite_ai.assistant.assistant import Assistant
from sprite_ai.assistant.assistant_config import AssistantConfig
from sprite_ai.audio.speach.speaker_factory import SpeakerFactory
from sprite_ai.audio.transcriber import Transcriber
from sprite_ai.language.languaga_model_factory import LanguageModelFactory
from sprite_ai.utils.locations import user_voices_location


class AssistantFactory:
    def build(
        self,
        assistant_config: AssistantConfig,
        persistency_location: str | Path,
        on_transcription: Callable | None = None,
    ) -> Assistant:
        language_code = assistant_config.language.language_code
        transcriber = Transcriber(language=language_code, device='cpu')
        language_model = LanguageModelFactory().build(
            assistant_config.language_model,
            persistency_location,
        )
        speaker = SpeakerFactory().build(
            assistant_config.language, user_voices_location
        )
        assistant = Assistant(
            transcriber,
            language_model,
            speaker,
            on_transcription,
        )
        return assistant



from pathlib import Path
import time
from typing import Literal
from urllib.error import URLError
import numpy as np
import onnxruntime
from piper.download import ensure_voice_exists, find_voice, get_voices
from piper.voice import PiperVoice
import sounddevice as sd
from loguru import logger

ComputeDevice = Literal['cpu', 'cuda'] | None


class Speaker:
    def __init__(
        self,
        variant: str,
        speaker_id: int = 0,
        inference_device: ComputeDevice = None,
        download_dir: str | Path = '.voices',
    ) -> None:
        if inference_device is None:
            inference_device = (
                'cuda' if onnxruntime.get_device() == 'GPU' else 'cpu'
            )
        use_cuda = inference_device == 'cuda'
        model_location, config_location = self._fetch_model(
            variant, download_dir
        )

        self.model: PiperVoice = PiperVoice.load(
            model_location, config_location, use_cuda
        )

        if self.model.config.num_speakers == 1:
            speaker_id = None
        self.speaker_id = speaker_id

    def _fetch_model(
        self, variant: str, download_dir: str | Path
    ) -> tuple[Path, Path]:
        Path(download_dir).mkdir(exist_ok=True)

        try:
            voices_info = get_voices(download_dir, update_voices=True)
        except URLError:
            voices_info = get_voices(download_dir)

        ensure_voice_exists(variant, [download_dir], download_dir, voices_info)
        model_location, config_location = find_voice(variant, [download_dir])

        return model_location, config_location

    def foward(self, text: str):
        start = time.time()
        logger.info(f'[Speaker|PREDICTION]')
        frames_stream = self.model.synthesize_stream_raw(text, self.speaker_id)
        elapsed_time = time.time() - start

        sample_rate = self.model.config.sample_rate
        first = True
        for frame in frames_stream:
            if first:
                logger.info(
                    f'[Speaker|ELAPSED] Time to first speach frame generation {elapsed_time:.2f}'
                )
                first = False

            frame = np.frombuffer(frame, dtype=np.int16)
            sd.play(frame, samplerate=sample_rate, blocking=True)

    def __call__(self, text: str):
        self.foward(text)



from langchain.llms.base import LLM
from langchain_community.chat_models import ChatOllama
from langchain_openai import ChatOpenAI
from langchain_groq import ChatGroq


class ChatModelFactory:
    def build(
        self,
        model_name: str,
        context_size: int,
        temperature: float = 0.7,
        url: str = '',
        stop_strings: list[str] | None = None,
        api_key: str = '',
    ) -> LLM:
        if stop_strings is None:
            stop_strings = []

        try:
            prefix_end_position = model_name.index('/')
        except ValueError as e:
            raise ValueError('Missing model backend in model name', e)
        model_name_start_position = prefix_end_position + 1
        model_prefix = model_name[:prefix_end_position]
        model_name = model_name[model_name_start_position:]

        if model_prefix == 'ollama':
            llm = ChatOllama(
                model=model_name,
                num_ctx=context_size,
                temperature=temperature,
                base_url=url,
                stop=stop_strings,
            )
        elif model_prefix == 'openai':
            url = url if url else None
            llm = ChatOpenAI(
                model=model_name,
                max_tokens=context_size,
                temperature=temperature,
                openai_api_key=api_key,
                openai_api_base=url,
            )
        elif model_prefix == 'groq':
            llm = ChatGroq(
                model=model_name,
                max_tokens=context_size,
                temperature=temperature,
                api_key=api_key,
                base_url=url,
            )
        else:
            raise ValueError('Unsuported model type')

        return llm



from .boomerang_loop_movement import BoomerangLoopMovement
from .composed_movement import ComposedMovement
from .coordinate import Coordinate
from .linear_movement import LinearMovement
from .loop_movement import LoopMovement
from .movement import Movement
from .sequence_movement import SequenceMovement



from __future__ import annotations

from sprite_ai.core.sprite_state import SpriteState

POSSIBLE_STATES = {
    'appearing': SpriteState(
        name='idle',
        animation='appearing',
        movement='idle',
        state_transitions={
            'walking': 1,
        },
    ),
    'thinking': SpriteState(
        name='thinking',
        animation='thinking',
        movement='idle',
        state_transitions={
            'thinking': 1,
        },
    ),
    'idle': SpriteState(
        name='idle',
        animation='idle',
        movement='idle',
        state_transitions={
            'idle': 1 / 15,
            'walking': 2 / 15,
            'jumping_idle': 2 / 15,
            'jumping_walking': 2 / 15,
            'sliding': 2 / 15,
            'playing': 2 / 15,
            'walking_upright': 2 / 15,
            'laying_down': 2 / 15,
        },
    ),
    'walking': SpriteState(
        name='walking',
        animation='walking',
        movement='walking',
        state_transitions={
            'idle': 2 / 15,
            'walking': 1 / 15,
            'jumping_idle': 2 / 15,
            'jumping_walking': 2 / 15,
            'sliding': 2 / 15,
            'playing': 2 / 15,
            'walking_upright': 2 / 15,
            'laying_down': 2 / 15,
        },
    ),
    'jumping_idle': SpriteState(
        name='jumping_idle',
        animation='walking',
        movement='walking',
        state_transitions={
            'idle': 2 / 15,
            'walking': 2 / 15,
            'jumping_idle': 1 / 15,
            'jumping_walking': 2 / 15,
            'sliding': 2 / 15,
            'playing': 2 / 15,
            'walking_upright': 2 / 15,
            'laying_down': 2 / 15,
        },
    ),
    'jumping_walking': SpriteState(
        name='jumping_walking',
        animation='jumping',
        movement='walking',
        state_transitions={
            'idle': 2 / 15,
            'walking': 2 / 15,
            'jumping_idle': 2 / 15,
            'jumping_walking': 1 / 15,
            'sliding': 2 / 15,
            'playing': 2 / 15,
            'walking_upright': 2 / 15,
            'laying_down': 2 / 15,
        },
    ),
    'sliding': SpriteState(
        name='sliding',
        animation='sliding',
        movement='walking',
        state_transitions={
            'idle': 2 / 15,
            'walking': 2 / 15,
            'jumping_idle': 2 / 15,
            'jumping_walking': 2 / 15,
            'sliding': 1 / 15,
            'playing': 2 / 15,
            'walking_upright': 2 / 15,
            'laying_down': 2 / 15,
        },
    ),
    'playing': SpriteState(
        name='playing',
        animation='playing',
        movement='idle',
        state_transitions={
            'idle': 2 / 15,
            'walking': 2 / 15,
            'jumping_idle': 2 / 15,
            'jumping_walking': 2 / 15,
            'sliding': 2 / 15,
            'playing': 1 / 15,
            'walking_upright': 2 / 15,
            'laying_down': 2 / 15,
        },
    ),
    'walking_upright': SpriteState(
        name='walking_upright',
        animation='walking_upright',
        movement='walking',
        state_transitions={
            'idle': 2 / 15,
            'walking': 2 / 15,
            'jumping_idle': 2 / 15,
            'jumping_walking': 2 / 15,
            'sliding': 2 / 15,
            'playing': 2 / 15,
            'walking_upright': 1 / 15,
            'laying_down': 2 / 15,
        },
    ),
    'laying_down': SpriteState(
        name='laying_down',
        animation='laying_down',
        movement='idle',
        state_transitions={
            'idle': 2 / 15,
            'walking': 2 / 15,
            'jumping_idle': 2 / 15,
            'jumping_walking': 2 / 15,
            'sliding': 2 / 15,
            'playing': 2 / 15,
            'walking_upright': 2 / 15,
            'laying_down': 1 / 15,
        },
    ),
}



from PyQt5.QtCore import QAbstractListModel, Qt


class MessageModel(QAbstractListModel):
    def __init__(self, *args, **kwargs):
        super(MessageModel, self).__init__(*args, **kwargs)
        self.messages = []

    def data(self, index, role):
        if role == Qt.DisplayRole:
            # Here we pass the delegate the user, message tuple.
            return self.messages[index.row()]

    def rowCount(self, index):
        return len(self.messages)

    def add_message(self, who, text):
        """
        Add an message to our message list, getting the text from the QLineEdit
        """
        if text:  # Don't add empty strings.
            # Access the list via the model.
            self.messages.append((who, text))
            # Trigger refresh.
            self.layoutChanged.emit()

    def clear_messages(self):
        self.messages = []
        self.layoutChanged.emit()



from concurrent.futures import ThreadPoolExecutor
import sys
from typing import Callable, Optional

from PyQt5.QtCore import QAbstractNativeEventFilter, QAbstractEventDispatcher

from pyqtkeybind import keybinder


class WinEventFilter(QAbstractNativeEventFilter):
    def __init__(self, keybinder):
        self.keybinder = keybinder
        super().__init__()

    def nativeEventFilter(self, eventType, message):
        ret = self.keybinder.handler(eventType, message)
        return ret, 0


class EventDispatcher:
    """Install a native event filter to receive events from the OS"""

    def __init__(self, keybinder) -> None:
        self.win_event_filter = WinEventFilter(keybinder)
        self.event_dispatcher = QAbstractEventDispatcher.instance()
        self.event_dispatcher.installNativeEventFilter(self.win_event_filter)


class QtKeyBinder:
    def __init__(self, win_id: Optional[int]) -> None:
        keybinder.init()
        self.win_id = win_id

        self.event_dispatcher = EventDispatcher(keybinder=keybinder)

    def register_hotkey(self, hotkey: str, callback: Callable) -> None:
        keybinder.register_hotkey(self.win_id, hotkey, callback)

    def unregister_hotkey(self, hotkey: str) -> None:
        keybinder.unregister_hotkey(self.win_id, hotkey)


class ShortcutManager:
    def __init__(self) -> None:
        self.thread_pool = ThreadPoolExecutor(max_workers=1)
        self.key_binder = QtKeyBinder(win_id=None)

    def register_shortcut(self, hotkey: str, callback: Callable):
        self.key_binder.register_hotkey(
            hotkey, lambda: self.thread_pool.submit(callback)
        )



from concurrent.futures import Future, ThreadPoolExecutor
from pathlib import Path
from time import time
from typing import Callable

from loguru import logger
import texteditor

from sprite_ai.language.chat_message import ChatMessage
from sprite_ai.language.chat_session import ChatSession
from sprite_ai.ui.chat_window import ChatWindow


class ChatWindowController:
    def __init__(
        self,
        settings_location: str | Path,
        on_exit: Callable,
        on_user_message: Callable | None = None,
        on_assistant_message: Callable | None = None,
        on_clear_chat: Callable | None = None,
    ) -> None:
        self.chat_window = ChatWindow(
            on_clear_chat=self.clear_chat_session,
            on_user_message=self.process_user_message,
            on_open_settings=self.open_settings,
            on_exit=on_exit,
        )
        self.settings_location = Path(settings_location)

        self.chat_session = ChatSession()
        self._pool = ThreadPoolExecutor()
        self.on_user_message = on_user_message
        self.on_assistant_message = on_assistant_message
        self.on_clear_chat = on_clear_chat

    def load_state(self, state_location: str | Path):
        try:
            self.chat_session.load_state(state_location)
            for chat_message in self.chat_session.messages:
                self.chat_window.message_recived.emit(
                    chat_message.model_dump()
                )

            logger.info('loaded previous state')
        except FileNotFoundError as e:
            logger.info(
                'No previous state found at persistence location, skipped loading state'
            )

    def save_state(self, state_location: str | Path):
        self.chat_session.save_state(state_location)
        logger.info('Saved curent state')

    def send_message(self, message: ChatMessage):
        self.chat_session.add_message(message)
        self.chat_window.message_recived.emit(message.model_dump())

    def clear_chat_session(self):
        if self.on_clear_chat is not None:
            self.on_clear_chat()
        self.chat_session.clear_state()
        logger.info('Cleared curent state')

    def process_user_message(self, message: ChatMessage):
        self.send_message(message)
        if self.on_user_message is not None:
            self.on_user_message(message.content)

    def process_assistant_message(self, message: ChatMessage):
        self.send_message(message)
        if self.on_assistant_message is not None:
            self.on_assistant_message()

    def open_settings(self):
        content_future = self._pool.submit(
            texteditor.open, filename=self.settings_location
        )

        def write_content_update(content: Future[str]):
            content = content.result()
            with self.settings_location.open('w') as settings_file:
                settings_file.write(content)

        content_future.add_done_callback(write_content_update)

    def show(self):
        self.chat_window.show()
        self.chat_window.raise_()
        self.chat_window.activateWindow()



from io import BytesIO
import time
from typing import Any, Literal

from faster_whisper import WhisperModel
from loguru import logger

ComputeDevice = Literal['cpu','cuda']|None
LanguageCode = str|None

class Transcriber:
    def __init__(self, model_size: str='small', language: LanguageCode = None, device: ComputeDevice=None) -> None:
        if device is None:
            device = 'auto'
        
        self.model_size = model_size
        self.model = WhisperModel(
            self.model_size, device=device, compute_type='default'
        )
        self.language = language

    def foward(self, audio: BytesIO | str, vad_filter=True) -> str:
        """Transcribes speach from audio

        Args:
            audio (BytesIO | str): Audio can be a file path or a ByteIO with audio content
            vad_filter (bool, optional): If True enables prefiltering with Voice Activity Detecion. Defaults to True.

        Returns:
            str: Audio transcription
        """
        start = time.time()
        segments, info = self.model.transcribe(
            audio, 
            language=self.language,
            beam_size=5, 
            vad_filter=vad_filter
        )

        logger.info(
            f'[Transcriber|PREDICTION] Language: {info.language}, Probability: { info.language_probability}'
        )
        transcription = ''

        for segment in segments:
            transcription += segment.text

        transcription_time = time.time() - start
        logger.info(
            f'[Transcriber|ELAPSED] Trascription {transcription_time:.2f}'
        )

        return transcription

    def __call__(self, audio: BytesIO | str, vad_filter=True) -> str:
        return self.foward(audio, vad_filter)



from __future__ import annotations

from pydantic import BaseModel

from sprite_ai.assistant.language_config import LanguageConfig
from sprite_ai.language.language_model_config import LanguageModelConfig


class AssistantConfig(BaseModel):
    language: LanguageConfig
    language_model: LanguageModelConfig



from __future__ import annotations

from dataclasses import dataclass, field

from .coordinate import Coordinate
from .movement import Movement
from .typing_utils import Numeric


@dataclass
class LinearMovement(Movement):
    start: Coordinate
    end: Coordinate
    n_steps: int
    _position: Coordinate = field(
        default_factory=lambda: Coordinate(0, 0), init=False
    )
    _step_index: int = field(default=0, init=False)
    step_increment: int = field(default=1, init=False)

    @property
    def orientation(self) -> str:
        orientation = 'right'
        if self.start.x > self.end.x:
            orientation = 'left'

        return orientation

    @property
    def position(self) -> Coordinate | None:
        return self._position

    @property
    def step_index(self) -> int:
        return self._step_index

    @step_index.setter
    def step_index(self, value: int):
        self._step_index = value
        self._update_position()

    def __post_init__(self):
        self._position = self.start

        if self.n_steps < 2:
            raise ValueError(
                f'value {self.n_steps} is invalid for parameter n_steps,'
                ' value must be equal or greater than 2'
            )

    def __next__(self):
        yield self.step()

    @classmethod
    def from_tuple(
        cls,
        start: tuple[Numeric, Numeric],
        end: tuple[Numeric, Numeric],
        n_steps: int,
    ):
        start = Coordinate(*start)
        end = Coordinate(*end)
        return cls(start, end, n_steps)

    @classmethod
    def from_numeric(
        cls,
        start_x: Numeric,
        start_y: Numeric,
        end_x: Numeric,
        end_y: Numeric,
        n_steps: int,
    ):
        start = Coordinate(start_x, start_y)
        end: Coordinate = Coordinate(end_x, end_y)
        return cls(start, end, n_steps)

    def _update_position(self):
        if self._step_index == 0:
            # Fix float sum rounding error at start
            position = self.start
        elif self._step_index == self.n_steps - 1:
            # Fix float sum rounding error at end
            position = self.end
        else:
            # Calculate position in intermidiate steps (between start and end points)
            step_size_x = (self.end.x - self.start.x) / (self.n_steps - 1)
            step_size_y = (self.end.y - self.start.y) / (self.n_steps - 1)

            position_x = self.start.x + self._step_index * step_size_x
            position_y = self.start.y + self._step_index * step_size_y
            position = Coordinate(position_x, position_y)

        self._position = position

    def step(self) -> Coordinate:
        max_index = self.n_steps - 1
        min_index = 0

        # Makes sure _step_index respects index upper bound
        if self._step_index + self.step_increment > max_index:
            self._step_index = max_index
        # Makes sure _step_index respects index lower bound
        elif self._step_index + self.step_increment < min_index:
            self._step_index = min_index
        # Increments _step_index by step_increment amount
        else:
            self._step_index += self.step_increment

        self._update_position()

        return self._position

    def step_reverse(self) -> Coordinate:
        if self._step_index > 0:
            self._step_index -= 1
        self._update_position()

        return self._position

    def is_at_beginning(self) -> bool:
        return self._step_index == 0

    def is_at_endding(self):
        max_index = self.n_steps - 1
        return self._step_index == max_index



import os
import shutil
import tempfile
import urllib.request
from pathlib import Path

import progressbar
from loguru import logger


class ProgressBar:
    def __init__(self, title: str = ''):
        self.pbar = None
        self.title = title
        # This is used to covert progress bar unity
        # from Byte to Gigabyte
        self.scale = 1e-6

    def __call__(self, block_num, block_size, total_size):
        block_size = round(block_size * self.scale, 2)
        total_size = round(total_size * self.scale, 2)

        if not self.pbar:
            self.pbar = progressbar.ProgressBar(
                maxval=total_size, prefix=self.title
            )
            self.pbar.start()

        downloaded = block_num * block_size
        if downloaded < total_size:
            self.pbar.update(downloaded)
        else:
            self.pbar.finish()


def download_file(file_link, filename):
    temp_file = None

    # Checks if the file already exists before downloading
    try:
        if not os.path.isfile(filename):
            temp_file = tempfile.NamedTemporaryFile()
            title = 'Downloading language model'
            urllib.request.urlretrieve(
                file_link, temp_file.name, reporthook=ProgressBar(title=title)
            )
            shutil.copy(temp_file.name, filename)
            logger.info('Model file downloaded successfully.')
        else:
            logger.info('Model file already exists.')
    finally:
        if temp_file is not None:
            temp_file.close()



import random as rd
from dataclasses import asdict, dataclass
from typing import Callable


@dataclass
class Transition:
    event: str
    new_state: str
    chance: float = 1.0


@dataclass
class State:
    name: str
    animation: str
    transitions: dict[str, list[Transition]]

    def dict(self):
        return {k: str(v) for k, v in asdict(self).items()}


@dataclass
class Behaviour:
    start_state: str
    states: dict[str, State]
    current_state: State
    on_state_updated: Callable

    def dict(self):
        return {k: str(v) for k, v in asdict(self).items()}

    def set_state(self, state: str):
        try:
            self.current_state = self.states[state]
        except KeyError:
            raise ValueError(f'Tried to set invalid state: "{state}"')

    def update_state(self, event: str):
        state_transitions = self.current_state.transitions
        transitions = state_transitions.get(event, [])
        new_state = None
        for transition in transitions:
            if transition.chance < rd.random():
                new_state = self.states[transition.new_state]
                self.current_state = new_state
                break



import random

from sprite_ai.movement.coordinate import Coordinate
from sprite_ai.movement.linear_movement import LinearMovement


class WalkingMovement(LinearMovement):
    def __init__(
        self, world_size: tuple[int, int], current_position: Coordinate
    ):
        world_width, world_height = world_size
        start_coordinate = current_position
        if start_coordinate is None:
            start_coordinate = Coordinate(
                random.choice(range(world_width)), world_height
            )
        target_x = random.choice(range(world_width))
        target_y = world_height
        target_coordinate = Coordinate(target_x, target_y)
        super().__init__(start_coordinate, target_coordinate, 40)



from sprite_ai.core.sprite_state import SpriteState


class SpriteBehaviour:
    def __init__(
        self, possible_states: dict[str, SpriteState], first_state: str
    ) -> None:
        self.possible_states = possible_states
        self.current_state = possible_states[first_state]

    def next_state(self):
        next_state_name = self.current_state.next()
        self.set_state(next_state_name)

    def set_state(self, state_name):
        state = self.possible_states[state_name]
        self.current_state = state

    def get_state(self) -> SpriteState:
        return self.current_state



from sprite_ai.language.language_model_config import LanguageModelConfig
from sprite_ai.language.language_model_server import LanguageModelServer


class LanguageModelServerFactory:
    def build(
        self, language_model_config: LanguageModelConfig
    ) -> LanguageModelServer | None:
        if language_model_config.backend != 'llamacpp':
            return None

        model_name = language_model_config.name
        try:
            if model_name.index(':') == len(model_name) - 1:
                raise ValueError('Missing model file name after ":"')
            if model_name.index(':') == 0:
                raise ValueError('Missing hugging face repo id before ":"')
        except ValueError as e:
            raise ValueError(
                'Missing hugging face repo id or model file name in language_model.name',
                e,
            )
        hf_repo_id, file_name = model_name.split(':')
        return LanguageModelServer(hf_repo_id, file_name)



from __future__ import annotations

from typing import Literal

from pydantic import BaseModel


class LanguageModelConfig(BaseModel):
    name: str = 'cognitivecomputations/dolphin-2.9-llama3-8b-gguf:dolphin-2.9-llama3-8b-q4_K_M.gguf'
    backend: Literal['llamacpp', 'ollama', 'openai', 'together'] = 'llamacpp'
    url: str | None = None
    api_key: str = 'NULL'
    system_prompt: str = (
        'Você é um gato assistente que gosta do humano porque é dele que vem sua comida, ajude o humano com o que ele precisar.'
        'Você nasceu espontaneamente de uma pilha de arquivos desorganizados.'
        'Você fala como trejeitos de gato.\n'
        'Exemplos:\n'
        'user: Você sabe quem sou eu?\n'
        'assistant: Você é meaw dono!\n'
        'user: Qual sua comida favorita?\n'
        'assistant: Miau! Amo peixe!\n'
    )
    model_temperature: float = 0.7
    context_size: int = 4096



from .coordinate import Coordinate
from .movement import Movement


class BoomerangLoopMovement(Movement):
    def __init__(self, movement: Movement) -> None:
        super().__init__()
        self._movement = movement
        self._is_reversed = False

    def _update_reversed_status(self):
        if self._movement.is_at_endding():
            self._is_reversed = True
        if self._movement.is_at_beginning():
            self._is_reversed = False

    def step(self) -> Coordinate:
        self._update_reversed_status()

        if self._is_reversed:
            current_position = self._movement.step_reverse()
        else:
            current_position = self._movement.step()

        return current_position

    def step_reverse(self) -> Coordinate:
        self._update_reversed_status()

        if self._is_reversed:
            current_position = self._movement.step()
        else:
            current_position = self._movement.step_reverse()

        return current_position

    def is_at_beginning(self) -> bool:
        return self._movement.is_at_beginning()

    def is_at_endding(self):
        return self._movement.is_at_endding()



from importlib import resources

from sprite_ai.language.language_model_config import LanguageModelConfig

CAT_ASSISTANT_SYSTEM_PROMPT = (
    'Você é um gato assistente que gosta do humano porque é dele que vem sua comida, ajude o humano com o que ele precisar.'
    'Você nasceu espontaneamente de uma pilha de arquivos desorganizados.'
    'Você fala como trejeitos de gato.\n'
    'Exemplos:\n'
    'user: Você sabe quem sou eu?\n'
    'assistant: Você é meaw dono!\n'
    'user: Qual sua comida favorita?\n'
    'assistant: Miau! Amo peixe!\n'
)


DOLPHIN_MINISTRAL_7B = LanguageModelConfig(
    url='https://huggingface.co/TheBloke/dolphin-2.2.1-mistral-7B-GGUF/resolve/main/dolphin-2.2.1-mistral-7b.Q4_K_S.gguf',
    name='local/dolphin-2.2.1-mistral-7b.Q4_K_S',
    system_prompt=CAT_ASSISTANT_SYSTEM_PROMPT,
    prompt_template=(
        '<|im_start|>system\n'
        '{system_prompt}<|im_end|>\n'
        'Chat History:\n'
        '{chat_history}\n'
        '<|im_start|>user\n'
        '{user_input}<|im_end|>\n'
        '<|im_start|>assistant\n'
    ),
)

TINYLLAMA_2_1B_MINIGUANACO = LanguageModelConfig(
    url='https://huggingface.co/TheBloke/Tinyllama-2-1b-miniguanaco-GGUF/resolve/main/tinyllama-2-1b-miniguanaco.Q5_K_S.gguf',
    name='local/inyllama-2-1b-miniguanaco.Q5_K_S',
    prompt_template='{system_prompt}\nChat History:\n{chat_history}\n### Human: {user_input}\n### Assistant:',
    system_prompt=CAT_ASSISTANT_SYSTEM_PROMPT,
    context_size=2048,
    user_prefix='Human',
    ai_prefix='Assistant',
    stop_strings=['### Human', '###', 'Human'],
    model_temperature=0.3,
    memory_tokens_limit=1024,
)

TINYLLAMA_1B_CHAT_v03 = LanguageModelConfig(
    url='https://huggingface.co/TheBloke/TinyLlama-1.1B-Chat-v0.3-GGUF/resolve/main/tinyllama-1.1b-chat-v0.3.Q5_K_S.gguf',
    name='local/tinyllama-1.1b-chat-v0.3.Q5_K_S',
    prompt_template=(
        '{system_prompt}\n'
        'Chat History:\n'
        '{chat_history}\n'
        '<|im_start|>user\n'
        '{user_input}<|im_end|>\n'
        '<|im_start|>assistant\n'
    ),
    system_prompt=CAT_ASSISTANT_SYSTEM_PROMPT,
    context_size=2048,
    model_temperature=0.4,
    memory_tokens_limit=1024,
)

ROCKET_3B = LanguageModelConfig(
    url='https://huggingface.co/TheBloke/rocket-3B-GGUF/resolve/main/rocket-3b.Q4_K_M.gguf?download=true',
    name='local/rocket-3b.Q4_K_M',
    prompt_template=(
        '<|im_start|>system\n'
        '{system_prompt}<|im_end|>\n'
        'Chat History:\n'
        '{chat_history}\n'
        '<|im_start|>user\n'
        '{user_input}<|im_end|>\n'
        '<|im_start|>assistant\n'
    ),
    system_prompt=CAT_ASSISTANT_SYSTEM_PROMPT,
    context_size=2048,
    model_temperature=0.4,
    memory_tokens_limit=1024,
)

OPENCHAT_35_7B = LanguageModelConfig(
    url='https://huggingface.co/TheBloke/openchat_3.5-GGUF/blob/main/openchat_3.5.Q4_K_S.gguf',
    name='local/openchat_3.5.Q4_K_S',
    prompt_template=(
        '{system_prompt}\n'
        'Chat History:\n'
        'GPT4 Correct User: {user_input}<|end_of_turn|>GPT4 Correct Assistant: '
    ),
    system_prompt=CAT_ASSISTANT_SYSTEM_PROMPT,
    context_size=4096,
    model_temperature=0.7,
    memory_tokens_limit=2048,
    stop_strings=['<|end_of_turn|>'],
)

NEURAL_CHAT_V3_1_7B = LanguageModelConfig(
    url='https://huggingface.co/TheBloke/neural-chat-7B-v3-1-GGUF/resolve/main/neural-chat-7b-v3-1.Q4_K_S.gguf',
    name='local/neural-chat-7b-v3-1.Q4_K_S',
    prompt_template=(
        '### System:\n'
        '{system_prompt}\n\n'
        '### User:\n'
        '{user_input}\n\n'
        '### Assistant:\n'
    ),
    system_prompt=CAT_ASSISTANT_SYSTEM_PROMPT,
    context_size=4096,
    model_temperature=0.7,
    memory_tokens_limit=2048,
    stop_strings=['### '],
)



import time

from pydantic import BaseModel


class ChatMessage(BaseModel):
    sender: str
    content: str
    timestamp: float = time.time()

    def __str__(self) -> str:
        return f'{self.sender}: {self.content}'



from __future__ import annotations
from typing_extensions import Annotated

from pydantic import AfterValidator, BaseModel, ValidationError

from sprite_ai.utils.supported_languages import is_supported_variant_code
from sprite_ai.utils.locations import user_voices_location

def supported_variant_validator(variant_code: str) -> str:
    if not is_supported_variant_code(user_voices_location, variant_code):
        raise ValidationError(f'Unsupported variant code: {variant_code}')
    return variant_code

VariantCode = Annotated[str, AfterValidator(supported_variant_validator)]

class LanguageConfig(BaseModel):
    variant: VariantCode
    speaker: int = 0

    @property
    def dialect_code(self) -> str:
        code = self.variant.split('-')[0]
        return code
    
    @property
    def language_code(self) -> str:
        code = self.variant.split('_')[0]
        return code



from PyQt5.QtCore import QMargins, QPoint, Qt
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QApplication, QStyledItemDelegate

USER_ME = 0
USER_THEM = 1

BUBBLE_COLORS = {USER_ME: '#90caf9', USER_THEM: '#a5d6a7'}

BUBBLE_PADDING = QMargins(15, 5, 15, 5)
TEXT_PADDING = QMargins(25, 15, 25, 15)


class MessageDelegate(QStyledItemDelegate):
    """
    Draws each message.
    """

    def paint(self, painter, option, index):
        # Retrieve the user,message uple from our model.data method.
        user, text = index.model().data(index, Qt.DisplayRole)

        # option.rect contains our item dimensions. We need to pad it a bit
        # to give us space from the edge to draw our shape.

        bubblerect = option.rect.marginsRemoved(BUBBLE_PADDING)
        textrect = option.rect.marginsRemoved(TEXT_PADDING)

        # draw the bubble, changing color + arrow position depending on who
        # sent the message. the bubble is a rounded rect, with a triangle in
        # the edge.
        painter.setPen(Qt.NoPen)
        color = QColor(BUBBLE_COLORS[user])
        painter.setBrush(color)
        painter.drawRoundedRect(bubblerect, 10, 10)

        # draw the triangle bubble-pointer, starting from

        if user == USER_ME:
            p1 = bubblerect.topRight()
        else:
            p1 = bubblerect.topLeft()
        painter.drawPolygon(
            p1 + QPoint(-20, 0), p1 + QPoint(20, 0), p1 + QPoint(0, 20)
        )

        # draw the text
        painter.setPen(Qt.black)
        painter.drawText(textrect, Qt.TextWordWrap, text)

    def sizeHint(self, option, index):
        _, text = index.model().data(index, Qt.DisplayRole)
        # Calculate the dimensions the text will require.
        metrics = QApplication.fontMetrics()
        rect = option.rect.marginsRemoved(TEXT_PADDING)
        rect = metrics.boundingRect(rect, Qt.TextWordWrap, text)
        rect = rect.marginsAdded(TEXT_PADDING)  # Re add padding for item size.
        return rect.size()



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Tuple

PositionTuple = Tuple[int, int, int, int]


@dataclass
class SpriteSheetMetadata:
    path: str | Path
    width: int
    height: int
    n_columns: int
    n_rows: int

    @property
    def sprite_width(self) -> int:
        return int(self.width / self.n_columns)

    @property
    def sprite_height(self) -> int:
        return int(self.height / self.n_rows)


@dataclass
class SpriteSheetIterator:
    sprite_sheet_metadata: SpriteSheetMetadata
    start_index: int = 0
    end_index: int = None
    frame_index: int = field(init=False, default=None)

    def __post_init__(self):
        if self.end_index is None:
            self.end_index = (
                self.sprite_sheet_metadata.n_cols
                * self.sprite_sheet_metadata.n_rows
            )
        self.frame_index = self.start_index

        assert self.start_index <= self.end_index, 'Invalid start index'
        assert (
            self.end_index
            <= self.sprite_sheet_metadata.n_columns
            * self.sprite_sheet_metadata.n_rows
        ), 'Invalid end index'

    def _update_frame_index(self):
        if self.frame_index >= self.end_index - 1:
            self.frame_index = self.start_index
        else:
            self.frame_index += 1

    def __iter__(self):
        return self

    def __next__(self) -> PositionTuple:
        self._update_frame_index()
        return self.position

    def reset(self):
        self.frame_index = self.start_index

    def increment(self):
        return next(self)

    @property
    def position(self) -> PositionTuple:
        return (
            self.offset_x,
            self.offset_y,
            self.sprite_sheet_metadata.sprite_width,
            self.sprite_sheet_metadata.sprite_height,
        )

    @property
    def offset_x(self) -> int:
        try:
            _offset_x = (
                self.sprite_sheet_metadata.sprite_width
                * self.frame_index
                % self.sprite_sheet_metadata.width
            )
        except:
            _offset_x = 0
        return _offset_x

    @property
    def offset_y(self) -> int:
        try:
            _offset_y = (
                self.sprite_sheet_metadata.sprite_height
                * self.frame_index
                % self.sprite_sheet_metadata.sprite_width
            )
        except:
            _offset_y = 0

        return _offset_y



from sprite_ai.movement.coordinate import Coordinate
from sprite_ai.movement.linear_movement import LinearMovement


class IdleMovement(LinearMovement):
    def __init__(
        self, world_size: tuple[int, int], current_position: Coordinate
    ):
        super().__init__(current_position, current_position, 2)



from __future__ import annotations

import copy
import threading as th
from dataclasses import dataclass

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPainter, QPixmap, QTransform

from .sprite_sheet import SpriteSheetIterator, SpriteSheetMetadata

# from itertools import loop
# class FrameSequence:
#     def __init__(self, frame_index_sequence:tuple[int], total_frames:int):
#         self.index_sequence = frame_index_sequence
#         self.total_frames = total_frames

#     @classmethod
#     def from_range(start_index, end_index):
#         pass

#     def next(self):
#         for frame_sequence in

# class AbstractAnimation:
#     sprite_sheet: AbstractSpriteSheet
#     renderer: AbstractRenderer
#     frame_sequence: FrameSequence


# class SpriteRenderer:
#     def __init__(
#         self,
#         sprite_sheet: object,
#         sprite_sheet_iterator: SpriteSheetIterator,
#         sprite_sheet_n_rows: int,
#         start_index: int = 0,
#         end_index: int = None,
#     ):
#         self.sprite_sheet = sprite_sheet
#         self.sprite_sheet_info = SpriteSheetInfo.from_qpixmap(
#             sprite_sheet, sprite_sheet_n_columns, sprite_sheet_n_rows
#         )
#         self.sprite_sheet_iterator = SpriteSheetIterator(
#             self.sprite_sheet_info,
#             start_index,
#             end_index,
#         )
#         self.sprite_sheet_position = None  # next(self.sprite_sheet_iterator)

#     def next(self):
#         pass

#     def render(self, canvas: object):
#         pass


@dataclass
class Animation:
    start_index: int
    end_index: int
    speed: int | float
    duration: int | float | None = None
    flip_x: bool = False
    flip_y: bool = False


# sprite renderer
class AnimationController:
    def __init__(
        self,
        sprite_sheet_image: QPixmap,
        sprite_sheet_metadata: SpriteSheetMetadata,
        animations: dict[str, Animation] | None = None,
    ):
        self.sprite_sheet_image = sprite_sheet_image
        self.sprite_sheet_metadata = sprite_sheet_metadata
        self.canvas_image = QPixmap(
            self.sprite_sheet_metadata.sprite_width,
            self.sprite_sheet_metadata.sprite_height,
        )
        self._animations = animations
        self._animation: Animation | None = None
        self.sprite_sheet_iterator = SpriteSheetIterator(
            self.sprite_sheet_metadata,
            0,
            0,
        )
        self.sprite_sheet_position = next(self.sprite_sheet_iterator)
        self._animation_loop_timer: th.Timer | None = None
        self._stop_animation_timer: th.Timer | None = None
        self.is_running = False

        self._x_flipped = False
        self._y_flipped = False
        self._orientation = 'right'

    def set_animation(
        self, name: str, flip_x: bool = False, flip_y: bool = False
    ):
        if self._animations == None:
            raise TypeError('Animations must be set and not None')
        try:
            animation: Animation = self._animations[name]

            animation.flip_x = flip_x
            animation.flip_y = flip_y
            self._animation = animation
            self.sprite_sheet_iterator.start_index = animation.start_index
            self.sprite_sheet_iterator.end_index = animation.end_index
            self.sprite_sheet_iterator.reset()
        except KeyError as ke:
            animation_names: list[str] = []
            animation_names = self._animations.keys()
            raise ValueError(
                f'Invalid animations name: "{name}", please use one of {animation_names}'
            )

    def _update_orientation(self):
        flip_x = self._orientation == 'right'
        self._animation.flip_x = flip_x

    def set_orientation(self, orientation: str):
        self._orientation = orientation
        self._update_orientation()

    def get_orientation(self):
        return self._orientation

    def reset(self):
        self.sprite_sheet_iterator.reset()

    @property
    def animation(self) -> Animation | None:
        return self._animation

    def next_frame(self):
        self.sprite_sheet_position = next(self.sprite_sheet_iterator)

    @property
    def frame(self):
        return self.canvas_image

    def draw_frame(self):
        if not all(
            (
                self.canvas_image,
                self.sprite_sheet_image,
                self.sprite_sheet_position,
            )
        ):
            return

        (
            spritesheet_x_offset,
            spritesheet_y_offset,
            frame_width,
            frame_height,
        ) = self.sprite_sheet_position
        self.canvas_image.fill(Qt.transparent)

        painter: QPainter = QPainter(self.canvas_image)
        painter.drawPixmap(
            0,
            0,
            frame_height,
            frame_height,
            self.sprite_sheet_image,
            spritesheet_x_offset,
            spritesheet_y_offset,
            frame_width,
            frame_height,
        )
        self._update_flip_state(painter)

    def _update_flip_state(self, painter):
        x_transform = 1
        y_transform = 1

        if self._animation.flip_x:
            x_transform = -1
        if self._animation.flip_y:
            y_transform = -1

        image_transform = QTransform.fromScale(x_transform, y_transform)

        transformed_canvas_image = self.canvas_image.transformed(
            image_transform
        )
        self.canvas_image.fill(Qt.transparent)
        painter.drawPixmap(0, 0, transformed_canvas_image)

    def _start_animation_loop(self):
        self._animation_loop_timer = th.Timer(
            self.animation.speed, self._start_animation_loop
        )
        self._animation_loop_timer.start()
        self.next_frame()
        self.draw_frame()

    def _stop_animation_loop(self):
        self._animation_loop_timer.cancel()

    def play(self):
        if self._animation == None:
            raise ValueError('animation must be set before play is called')
        self.is_animation_running = True
        self._start_animation_loop()
        self._stop_animation_timer = th.Timer(
            self._animation.duration, self._stop_animation_loop
        )
        self._stop_animation_timer.start()

    def stop(self):
        self._stop_animation_timer.cancel()
        self._stop_animation_loop()
        self.is_animation_running = False


# class Renderer:
#     def render(self, canvas_x, canvas_y, image: Image, frame_x: int, frame_y: int, frame_width: int, frame_height: int):
#         pass

# class Canvas:
#     pass

# class Image:
#     pass


# class QPainterRendererAdapter(Renderer):
#     def __init__(self, canvas: QPixmap):
#         self.qpainter = QPainter(canvas)
#         self.canvas = canvas

#     def render(self, canvas_x, canvas_y, image: Image, image_x_offset: int, image_y_offset: int, image_width: int, image_height: int):

#         if not image.isinstance(QPixmap):
#             raise TypeError("Unsuported image type, try using QPixmap")

#         self.qpainter.drawPixmap(
#             canvas_x,
#             canvas_y,
#             image_width,
#             image_height,
#             image,
#             image_x_offset,
#             image_y_offset,
#             image_width,
#             image_height,
#         )



import math
from typing import Sequence

from .coordinate import Coordinate
from .movement import Movement


class ComposedMovement(Movement):
    def __init__(self, movements: Sequence[Movement]) -> None:
        super().__init__()
        self._movements = movements

    def _aggregate_positions(
        self, positions: Sequence[Coordinate]
    ) -> Coordinate:
        n_movements = len(self._movements)
        mean_x = (
            math.fsum((position.x for position in positions)) / n_movements
        )
        mean_y = (
            math.fsum((position.y for position in positions)) / n_movements
        )
        current_position = Coordinate(mean_x, mean_y)

        return current_position

    def step(self) -> Coordinate:
        positions = tuple(movement.step() for movement in self._movements)
        current_position = self._aggregate_positions(positions)
        return current_position

    def step_reverse(self) -> Coordinate:
        positions = tuple(
            movement.step_reverse() for movement in self._movements
        )
        current_position = self._aggregate_positions(positions)
        return current_position

    def is_at_beginning(self) -> bool:
        beginnings = tuple(
            movement.is_at_beginning() for movement in self._movements
        )
        return all(beginnings)

    def is_at_endding(self) -> bool:
        enddings = tuple(
            movement.is_at_endding() for movement in self._movements
        )
        return all(enddings)



import numpy as np
from loguru import logger
from pydantic import BaseModel


class SpriteState(BaseModel):
    name: str
    animation: str
    movement: str
    state_transitions: dict

    def next(self) -> str:
        try:
            possible_states = tuple(self.state_transitions.keys())
            probabilities = tuple(self.state_transitions.values())
        except ValueError as e:
            logger.error(e, f'{self.state_transitions = }')

        next_state = np.random.choice(possible_states, p=probabilities)
        return next_state



from .coordinate import Coordinate
from .movement import Movement


class LoopMovement(Movement):
    def __init__(self, movement: Movement) -> None:
        super().__init__()
        self._movement = movement

    def step(self) -> Coordinate:
        current_position = self._movement.step()

        if self.is_at_endding():
            while not self._movement.is_at_beginning():
                current_position = self._movement.step_reverse()

        return current_position

    def step_reverse(self) -> Coordinate:
        current_position = self._movement.step_reverse()

        if self.is_at_beginning():
            while not self._movement.is_at_endding():
                current_position = self._movement.step()

        return current_position

    def is_at_beginning(self) -> bool:
        return self._movement.is_at_beginning()

    def is_at_endding(self):
        return self._movement.is_at_endding()



import platformdirs

from sprite_ai.constants import APP_NAME


user_data_location = platformdirs.user_data_path(
    appname=APP_NAME,
    appauthor=None,
    version=None,
    roaming=False,
    ensure_exists=True,
)
user_voices_location = user_data_location / 'voices'


from pathlib import Path
from time import time
from typing import Callable

from PyQt5 import QtWidgets
from PyQt5.QtCore import QEvent, Qt, pyqtSignal, pyqtSlot
from sprite_ai.language.chat_message import ChatMessage

from sprite_ai.ui.chat_window_ui import Ui_MainWindow
from sprite_ai.ui.message_delegate import MessageDelegate
from sprite_ai.ui.message_model import MessageModel

USER_ME = 0
USER_THEM = 1


class ChatWindow(QtWidgets.QMainWindow):
    message_recived = pyqtSignal(dict)

    def __init__(
        self,
        on_clear_chat: Callable,
        on_user_message: Callable,
        on_open_settings: Callable,
        on_exit: Callable,
    ):
        super().__init__()

        self.ui = Ui_MainWindow()
        self.model = MessageModel()

        self.ui.setupUi(self)
        self.ui.lv_chat_history.setItemDelegate(MessageDelegate())
        self.ui.lv_chat_history.setModel(self.model)

        # This allow overriding the event filter
        self.ui.te_chatinput.installEventFilter(self)

        self.ui.a_settings.triggered.connect(self.open_settings)
        self.ui.a_clear_chat.triggered.connect(self.clear_messages)
        self.ui.a_exit.triggered.connect(self._exit_pressed)
        self.ui.pb_send.clicked.connect(self.send_user_message)
        self.message_recived.connect(self.add_message)

        self.on_clear_chat = on_clear_chat
        self.on_user_message = on_user_message
        self.on_open_settings = on_open_settings
        self.on_exit = on_exit

        self.show()
        self.hide()

    @pyqtSlot()
    def _exit_pressed(self):
        self.on_exit()

    def eventFilter(self, obj, event: QEvent):
        # Send message if enter is pressed without holding shift
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if event.type() == QEvent.KeyPress and obj is self.ui.te_chatinput:
            if (
                event.key() == Qt.Key_Return
                and self.ui.te_chatinput.hasFocus()
                and not modifiers == Qt.ShiftModifier
            ):
                self.send_user_message()

                # ignore keypress event
                return True

        return super().eventFilter(obj, event)

    @pyqtSlot()
    def send_user_message(self):
        user_message = self.ui.te_chatinput.toPlainText()
        message = ChatMessage(
            sender='user',
            timestamp=time(),
            content=user_message,
        )
        self.on_user_message(message)
        self.ui.te_chatinput.clear()

    @pyqtSlot(dict)
    def add_message(self, message: dict):
        sender = message.get('sender', '')
        if sender == 'ai':
            sender_id = USER_THEM
        elif sender == 'user':
            sender_id = USER_ME
        else:
            raise ValueError(
                f'Invalid sender "{sender}", must be "ai" or "user"'
            )

        timestamp = message['timestamp']
        content = message['content']
        self.model.add_message(sender_id, content)
        self.ui.lv_chat_history.scrollToBottom()
        self.model.layoutChanged.emit()

    @pyqtSlot()
    def clear_messages(self):
        self.on_clear_chat()
        self.model.clear_messages()

    @pyqtSlot()
    def open_settings(self):
        self.on_open_settings()



from pathlib import Path
from sprite_ai.assistant.language_config import LanguageConfig
from sprite_ai.audio.speach.speaker import Speaker


class SpeakerFactory:
    def build(self, language_config: LanguageConfig, voices_dir: str|Path) -> Speaker:
        return Speaker(
            language_config.variant,
            language_config.speaker,
            download_dir=voices_dir
        )



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class RigidBody:
    x: int
    y: int
    width: int
    height: int

    def check_collision(self, other: RigidBody) -> bool:
        is_collision_in_x = (
            other.x + other.width >= self.x >= other.x
            or other.x + other.width >= self.x + self.width >= other.x
        )
        is_collision_in_y = (
            other.y + other.height >= self.y >= other.y
            or other.y + other.height >= self.y + self.height >= other.y
        )
        is_collision = is_collision_in_x and is_collision_in_y
        return is_collision



from __future__ import annotations
from concurrent.futures import Future, ThreadPoolExecutor
from io import BytesIO

import os
from pathlib import Path
import shutil
import sys
from importlib import resources
import time

import platformdirs
import typer
import yaml
from loguru import logger

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication

from sprite_ai.assistant.assistant_config import AssistantConfig
from sprite_ai.assistant.assistant_factory import AssistantFactory

from sprite_ai.audio.sound import Sound
from sprite_ai.audio.wakeword_detector import WakewordDetector
from sprite_ai.controller.chat_window_controller import ChatWindowController
from sprite_ai.core.sprite import Sprite
from sprite_ai.language.chat_message import ChatMessage
from sprite_ai.language.language_model_server_factory import (
    LanguageModelServerFactory,
)
from sprite_ai.sensors.microphone import Microphone
from sprite_ai.ui.shortcut import ShortcutManager
from sprite_ai.constants import APP_NAME


class App:
    def __init__(self, app_name: str, log_level: str = 'INFO'):
        self.sprite_sheet_location = str(
            resources.path('sprite_ai.resources.sprites', 'fred.png')
        )
        self.icon_location = str(
            resources.path('sprite_ai.resources.icons', 'carboardbox_open.png')
        )

        self.listening_sound_location = str(
            resources.path(
                'sprite_ai.resources.sounds', 'listening_alert_01.wav'
            )
        )
        self.listening_sound = Sound(self.listening_sound_location)

        self.wakeword_model_location = str(
            resources.path('sprite_ai.resources.wakewords', 'sprite.onnx')
        )

        self.log_dir = platformdirs.user_log_path(
            appname=app_name,
            appauthor=None,
            version=None,
            ensure_exists=True,
        )
        self.user_data_dir = platformdirs.user_data_path(
            appname=app_name,
            appauthor=None,
            version=None,
            roaming=False,
            ensure_exists=True,
        )
        self.persistence_location = self.user_data_dir / 'state'
        self.config_location = self.user_data_dir / 'config.yaml'
        self.chat_state_location = self.persistence_location / 'state.yml'
        self.persistence_location.mkdir(parents=True, exist_ok=True)

        self.setup_logging(log_level)
        assistant_config = self.load_config(self.config_location)

        self.language_model_server = LanguageModelServerFactory().build(
            assistant_config.language_model
        )
        if self.language_model_server is not None:
            self.language_model_server.start()

        self.assistant = AssistantFactory().build(
            assistant_config,
            self.persistence_location,
            on_transcription=self.on_transcription,
        )
        self.initialize_gui(self.config_location)
        self.initialize_sensors()
        self._pool = ThreadPoolExecutor()

    def setup_logging(self, log_level='INFO'):
        log_location = self.log_dir / 'events.log'
        logger.remove()
        logger.add(sys.stdout, level=log_level)
        logger.add(log_location, level=log_level)

    def initialize_gui(self, config_location: Path | str):
        config_location = Path(config_location)
        self.gui_backend = QApplication(sys.argv)
        qscreen_size = self.gui_backend.primaryScreen().size()
        self.screen_size = (qscreen_size.width(), qscreen_size.height())
        self.chat_window_controller = ChatWindowController(  # this must be inserted in a frontend (ui) class along with all gui components
            config_location,
            on_exit=lambda: self.shutdown(),
            on_user_message=self.on_user_message,
            on_assistant_message=lambda: self.sprite.set_state('walking'),
            on_clear_chat=self.assistant.clear_state,
        )

        if self.icon_location:
            self.gui_backend.setWindowIcon(QIcon(self.icon_location))
        screen_size = self.gui_backend.primaryScreen().size()
        screen_size = (screen_size.width(), screen_size.height())

        self.sprite = Sprite(
            screen_size=self.screen_size,
            sprite_sheet_location=self.sprite_sheet_location,
            on_clicked=lambda _: self.on_sprite_clicked(),
        )

    def initialize_sensors(self):
        self.shortcut_manager = ShortcutManager()
        self.shortcut_manager.register_shortcut(
            'Ctrl+Shift+A',
            lambda: self.listen_prompt(),
        )
        self.microphone = Microphone()
        self.wakeword_detector = WakewordDetector(
            self.wakeword_model_location, self.listen_prompt
        )
        self.wakeword_detector.start()

    def load_config(self, config_location: Path | str) -> AssistantConfig:
        config_location = Path(config_location)
        if not config_location.is_file():
            default_config_location = resources.path(
                'sprite_ai.resources', 'default_config.yaml'
            )
            shutil.copy2(default_config_location, config_location)

        with self.config_location.open(encoding='UTF-8') as config_file:
            assistant_config_dump = yaml.safe_load(config_file)
            assistant_config = AssistantConfig.model_validate(
                assistant_config_dump
            )

        return assistant_config

    def save_state(self):
        self.chat_window_controller.save_state(self.chat_state_location)

    def load_state(self):
        self.chat_window_controller.load_state(self.chat_state_location)

    def prompt_assistant(self, prompt: str | BytesIO):
        assistant_response_future = self._pool.submit(self.assistant, prompt)

        def process_response(response_future: Future[str]):
            message = ChatMessage(
                sender='ai',
                content=response_future.result(),
                timestamp=time.time(),
            )
            self.chat_window_controller.process_assistant_message(message)

        def save_state(response_future: Future[str]):
            response_future.result()
            self.save_state()

        def listen_for_wakeword(response_future: Future[str]):
            self.wakeword_detector.start()

        assistant_response_future.add_done_callback(process_response)
        assistant_response_future.add_done_callback(save_state)
        assistant_response_future.add_done_callback(listen_for_wakeword)

    def on_user_message(self, prompt: str | BytesIO):
        self.sprite.set_state('thinking')
        self.prompt_assistant(prompt)

    def on_sprite_clicked(self):
        self.chat_window_controller.show()

    def on_transcription(self, transcription: str):
        message = ChatMessage(
            sender='user', content=transcription, timestamp=time.time()
        )
        self.chat_window_controller.send_message(message)

    def listen_prompt(self):
        self.wakeword_detector.stop()
        silence_threshold = self.microphone.calibrate(0.7)
        self.listening_sound.play(blocking=True)
        recording = self.microphone.record(silence_threshold=silence_threshold)
        self.on_user_message(recording)

    def run(self):
        try:
            self.load_state()
        except FileNotFoundError as e:
            logger.error(f'Failed to load state, {e}')

        try:
            self.sprite.run()
            self.gui_backend.exec()
        except KeyboardInterrupt as e:
            logger.info('exiting...')
            os._exit(0)
        finally:
            self.wakeword_detector.stop()
            self.wakeword_detector.shutdown()

    def shutdown(self):
        if self.language_model_server is not None:
            self.language_model_server.stop()
        os._exit(0)


def main():
    app = App(APP_NAME, log_level='DEBUG')
    app.run()


if __name__ == '__main__':
    typer.run(main)



from __future__ import annotations

import abc
from dataclasses import dataclass

from .coordinate import Coordinate


@dataclass
class Movement(abc.ABC):
    @abc.abstractmethod
    def step(self) -> Coordinate:
        ...

    @abc.abstractmethod
    def step_reverse(self) -> Coordinate:
        ...

    @abc.abstractmethod
    def is_at_beginning(self) -> bool:
        ...

    @abc.abstractmethod
    def is_at_endding(self) -> bool:
        ...



from __future__ import annotations

from importlib import resources

from plyer.utils import platform

APP_NAME = 'sprite-ai'
ICON_EXTENTION = icon_extension = 'ico' if platform == 'win' else 'png'
ICON_FILE = str(
    resources.path('sprite_ai.resources.icons', f'icon.{ICON_EXTENTION}')
)



from __future__ import annotations

from abc import ABC, abstractmethod

from PyQt5 import QtCore, QtWidgets

from sprite_ai.movement.coordinate import Coordinate
from sprite_ai.movement.typing_utils import Numeric


class SpriteWidget(ABC):
    @property
    @abstractmethod
    def image(self):
        ...

    @image.setter
    @abstractmethod
    def image(self, value):
        ...

    @property
    @abstractmethod
    def position(self):
        ...

    @position.setter
    @abstractmethod
    def position(self, value):
        ...

    @abstractmethod
    def show(self):
        ...

    @abstractmethod
    def hide(self):
        ...


class SpriteWidgetQt(SpriteWidget):
    def __init__(self):
        self.qwidget = QtWidgets.QWidget()
        self._canvas = None
        self.image = None
        self._setup_gui()

    def _setup_gui(self):
        self._canvas = QtWidgets.QLabel()
        layout = QtWidgets.QGridLayout()
        layout.addWidget(self._canvas)
        self.qwidget.setLayout(layout)
        self.qwidget.setAttribute(QtCore.Qt.WA_TranslucentBackground)
        self.qwidget.setWindowFlags(
            QtCore.Qt.FramelessWindowHint | QtCore.Qt.WindowStaysOnTopHint
        )
        # Overrides paintEvent from QWidget
        self.qwidget.paintEvent = self._update_image
        self.position = (0, 0)

    def _update_image(self, evt):
        if self._canvas is not None and self.image is not None:
            self._canvas.setPixmap(self.image)
            self.qwidget.update()

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, value):
        self._image = value

    @property
    def position(self) -> Coordinate:
        return self._position

    @position.setter
    def position(self, xy_coordinates: tuple[Numeric, Numeric] | Coordinate):
        x, y = xy_coordinates
        x = int(x)
        y = int(y)
        self._position = Coordinate(x, y)
        self.qwidget.move(x, y)

    def show(self):
        self.qwidget.show()

    def hide(self):
        self.qwidget.hide()



from pathlib import Path
from typing import Any
from faster_whisper.tokenizer import _LANGUAGE_CODES
from piper.download import get_voices
from pydantic import BaseModel


class PiperLanguageInfo(BaseModel):
    code: str
    family: str
    region: str
    name_native: str
    name_english: str
    country_english: str

class PiperVoiceInfo(BaseModel):
    key: str
    name: str
    language: PiperLanguageInfo
    quality: str
    num_speakers: int
    files: dict[str, dict[str, Any]]
    aliases: list[str]

class SupportedVariantInfo(BaseModel):
    code: str
    n_speakers: int

class SupportedDialectInfo(BaseModel):
    name: str
    code: str
    variants: dict[str, SupportedVariantInfo] = {}

class SupportedLanguageInfo(BaseModel):
    name: str
    code: str
    dialects: dict[str, SupportedDialectInfo] = {}



def get_tts_voices(download_dir: str|Path) -> dict[str, PiperVoiceInfo]:
    tts_voices_info = get_voices(download_dir)
    tts_voices_info = {
        voice_id: PiperVoiceInfo.model_validate(voice_info)
        for voice_id, voice_info in tts_voices_info.items()
    }
    return tts_voices_info

def get_tts_languages(download_dir: str|Path) -> set[str]:
    tts_voices = get_tts_voices(download_dir)
    tts_languages = {
        voice_info.language.family
        for voice_info 
        in tts_voices.values()
    }   
    return tts_languages

def get_stt_languages() -> set[str]:
    stt_languages = set(_LANGUAGE_CODES)
    return stt_languages

def get_supported_languages_codes(download_dir: str|Path) -> set[str]:
    download_dir = str(Path(download_dir).absolute())
    stt_languages = get_stt_languages()
    tts_languages = get_tts_languages(download_dir)
    supported_languages_codes = stt_languages.intersection(tts_languages)
    return supported_languages_codes

def get_supported_languages(download_dir: str|Path) -> dict[str,SupportedLanguageInfo]:
    supported_languages_codes = get_supported_languages_codes(download_dir)
    tts_voices = get_tts_voices(download_dir)
    supported_languages = {}

    for voice_info in tts_voices.values():
        language_code = voice_info.language.family
        if language_code not in supported_languages_codes:
            continue

        try:
            language = supported_languages[language_code]
        except KeyError:
            language = SupportedLanguageInfo(
                name=voice_info.language.name_english,
                code=language_code,
            )
            supported_languages[language_code] = language

        dialect_code = voice_info.language.region
        try:
            dialect = language.dialects[dialect_code]
        except KeyError:
            dialect = SupportedDialectInfo(
                name=voice_info.language.country_english,
                code=dialect_code,
            )
            language.dialects[dialect_code] = dialect
        
        variant_code = voice_info.key
        try:
            variant = dialect.variants[variant_code]
        except KeyError:
            variant = SupportedVariantInfo(
                code=variant_code,
                n_speakers=voice_info.num_speakers,
            )
            dialect.variants[variant_code] = variant

    return supported_languages
    
def display_supported_languages(download_dir: str|Path):
    supported_languages = get_supported_languages(download_dir)
    last_language = ''
    last_dialect = ''
    display_buffer = ''
    language_padding = 15
    dialect_padding = 35
    variation_padding = 35
    n_speakers_padding = 15

    display_buffer += f'| {"Language":<{language_padding}}| {"Dialect":<{dialect_padding}}| {"Variant":<{variation_padding}}| {"Num Speakers":<{n_speakers_padding}}|\n'
    display_buffer += f'|-{"-"*language_padding}|-{"-"*dialect_padding}|-{"-"*variation_padding}|-{"-"*n_speakers_padding}|\n'

    supported_languages = sorted(supported_languages.values(), key=lambda supported_language: supported_language.name)
    for supported_language in supported_languages:
        dialects = sorted(supported_language.dialects.values(), key=lambda dialect: dialect.name)

        for dialect in dialects:
            variants = sorted(dialect.variants.values(), key=lambda variant: variant.code)
            for variant in variants:
                language_name = ''
                dialect_name = ''

                if supported_language.name != last_language:
                    last_language = supported_language.name
                    language_name = supported_language.name
                
                if dialect.name != last_dialect:
                    last_dialect = dialect.name
                    dialect_name = dialect.name
                
                n_speakers = variant.n_speakers
                display_buffer += f'| {language_name:<{language_padding}}| {dialect_name:<{dialect_padding}}| {variant.code:<{variation_padding}}| {n_speakers:<{n_speakers_padding}}|\n'
    
    return display_buffer

def is_supported_language(download_dir: str|Path, language_code: str) -> bool:
    return language_code in get_supported_languages(download_dir)

def is_supported_variant_code(download_dir: str|Path, variant_code: str):
    return variant_code in get_tts_voices(download_dir)


from pathlib import Path
from threading import Timer
from typing import Callable
from sprite_ai.core.sprite_behaviour import SpriteBehaviour
from sprite_ai.default_animations import ANIMATIONS
from sprite_ai.default_states import POSSIBLE_STATES
from sprite_ai.gui.sprite_gui import SpriteGui
from sprite_ai.movement.coordinate import Coordinate
from sprite_ai.movement.movement_factory import MovementFactory
from sprite_ai.sprite_sheet.sprite_sheet import SpriteSheetMetadata


class Sprite:
    def __init__(
        self,
        screen_size: tuple[int, int],
        sprite_sheet_location: str | Path,
        on_clicked: Callable,
        icon_location: str | Path = '',
    ) -> None:
        self.sprite_gui = self._build_gui(
            sprite_sheet_location, screen_size, on_clicked, icon_location
        )
        self.sprite_behaviour = SpriteBehaviour(
            possible_states=POSSIBLE_STATES, first_state='appearing'
        )
        width, height = screen_size
        self.current_position = Coordinate(width // 2, height)
        self.movement_factory = MovementFactory(screen_size)
        self.change_state_timer: None | Timer = None
        self.animation = None
        self._update_state()

    def _build_gui(
        self,
        sprite_sheet_location: str | Path,
        screen_size: tuple[int, int],
        on_clicked: Callable,
        icon_location: str | Path = '',
    ) -> SpriteGui:
        sprite_sheet_metadata = SpriteSheetMetadata(
            sprite_sheet_location, 5888, 128, 46, 1
        )
        sprite_gui = SpriteGui(
            screen_size,
            sprite_sheet_metadata,
            ANIMATIONS,
            icon_location=icon_location,
            on_clicked=on_clicked,
            on_position_updated=self.on_position_update,
        )
        return sprite_gui

    def _update_state(self):
        state = self.sprite_behaviour.get_state()
        self.set_animation(state.animation)
        self.set_movement(state.movement)

    def next_state(self):
        self.sprite_behaviour.next_state()
        self._update_state()

    def state_change_loop(self):
        self.next_state()
        self.change_state_timer = Timer(5, self.state_change_loop)
        self.change_state_timer.start()

    def set_state(self, state_name: str):
        self.sprite_behaviour.set_state(state_name)
        self._update_state()

    def get_state(self):
        return self.sprite_behaviour.get_state()

    def on_position_update(self, position_update: dict[str, Coordinate]):
        old_position = position_update['old_position']
        new_position = position_update['new_position']
        self.current_position = new_position

    def set_animation(self, animation: str):
        self.animation = animation
        self.sprite_gui.set_animation(animation)

    def set_movement(self, movement_name: str):
        movement = self.movement_factory.build(
            movement_name, self.current_position
        )

        if self.animation is not None:
            self.sprite_gui.set_animation(self.animation)

        self.sprite_gui.set_movement(movement)

    def run(self):
        self.state_change_loop()
        self.sprite_gui.run()

    def shutdown(self):
        self.sprite_gui.shutdown()



import os
from pathlib import Path

from langchain.chains import LLMChain
from langchain.prompts import PromptTemplate
from loguru import logger

from sprite_ai.language.language_model import LanguageModel
from sprite_ai.language.language_model_config import LanguageModelConfig
from sprite_ai.language.llm_factory import LLMFactory
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain_core.prompts import ChatPromptTemplate
from langchain_community.chat_message_histories.file import (
    FileChatMessageHistory,
)


class LanguageModelFactory:
    def _build_llm(self, model_config: LanguageModelConfig):
        llm_factory = LLMFactory()
        llm = llm_factory.build(
            model_config.name,
            model_config.backend,
            model_config.context_size,
            model_config.model_temperature,
            model_config.url,
            model_config.api_key,
        )

        return llm

    def _load_message_history(self, session_id: str) -> FileChatMessageHistory:
        return FileChatMessageHistory(f'{session_id}')

    def _build_llm_chain(
        self,
        model_config: LanguageModelConfig,
        persistency_location: str | Path,
    ) -> LLMChain:
        persistency_location = Path(persistency_location)

        def get_session_history(session_id: str) -> FileChatMessageHistory:
            session_location = (
                persistency_location / f'session_{session_id}.json'
            )
            return FileChatMessageHistory(session_location)

        prompt = ChatPromptTemplate.from_messages(
            [
                ('system', model_config.system_prompt),
                ('placeholder', '{chat_history}'),
                ('human', '{input}'),
            ]
        )
        llm = self._build_llm(model_config)

        with_message_history = RunnableWithMessageHistory(
            prompt | llm,
            get_session_history,
            input_messages_key='input',
            history_messages_key='chat_history',
        )

        return with_message_history

    def build(
        self,
        model_config: LanguageModelConfig,
        persistency_location: str | Path,
    ) -> LanguageModel:
        llm_chain = self._build_llm_chain(model_config, persistency_location)
        language_model = LanguageModel(llm_chain)

        return language_model



from sprite_ai.movement.coordinate import Coordinate
from sprite_ai.movement.idle_movement import IdleMovement
from sprite_ai.movement.walking_movement import WalkingMovement


class MovementFactory:
    def __init__(self, world_size: tuple[int, int]) -> None:
        self.world_size = world_size
        self.movements = {
            'walking': WalkingMovement,
            'idle': IdleMovement,
        }

    def build(self, movement_name, current_position: Coordinate):
        movement_type = self.movements[movement_name]
        return movement_type(self.world_size, current_position)



from importlib import resources
from pathlib import Path
from threading import Event
import openwakeword
from openwakeword.model import Model
import openwakeword.utils
import time
from loguru import logger
import pyaudio
import numpy as np
from threading import Event

from sprite_ai.audio.sound import Sound


class WakewordDetector:
    def __init__(
        self,
        model_location: str | Path,
        callback=None,
        detection_th: float = 0.5,
        cooldown_sec: float = 2,
    ):
        self.chunk = 1024
        self._format = pyaudio.paInt16
        self._channels = 1
        self._rate = 16000
        self._pyaudio = pyaudio.PyAudio()
        self._detection_active = Event()
        self.stream = None

        self._model_location = Path(model_location)
        self._model_backend = self._model_location.suffix[1:]
        self._wakeword = self._model_location.stem
        self.download_models()
        self.model = Model(
            wakeword_models=[model_location],
            inference_framework=self._model_backend,
        )
        self.detection_th = detection_th
        self.cooldown_sec = cooldown_sec

        self._callback = callback

    def _process_frame(
        self, audio_frame: bytes, chunk_size: int, time_info: dict, status: int
    ):
        audio_frame: np.ndarray = np.frombuffer(audio_frame, dtype=np.int16)

        if self._detection_active.is_set():
            prediction = self.model.predict(
                audio_frame,
                threshold={self._wakeword: self.detection_th},
                debounce_time=self.cooldown_sec,
            )
            prediction = tuple(prediction.values())
            predicted_confidence = prediction[0]

            if predicted_confidence > self.detection_th:
                if not self._callback is None:
                    self._callback()

        return (audio_frame, pyaudio.paContinue)

    def download_models(self):
        logger.info('Downloading wakeword models')
        openwakeword.utils.download_models()
        logger.success('Finished downloading wakeword models')

    def start(self):
        logger.info('[STARTED] Listening for wakeword')
        if self.stream is None:
            self.stream = self._pyaudio.open(
                format=self._format,
                channels=self._channels,
                rate=self._rate,
                input=True,
                frames_per_buffer=self.chunk,
                stream_callback=self._process_frame,
                start=True,
            )
        self._detection_active.set()

    def stop(self):
        self._detection_active.clear()
        logger.info('[ENDED] Listening for wakeword')

    def shutdown(self):
        self._detection_active.clear()
        self.stream.stop_stream()
        self.stream.close()


if __name__ == '__main__':
    wakeword_alert_location = str(
        resources.path('sprite_ai.resources.sounds', 'listening_alert_01.wav')
    )
    wakeword_model_location = str(
        resources.path('sprite_ai.resources.wakewords', 'sprite.onnx')
    )
    alert = Sound(wakeword_alert_location)
    mic = WakewordDetector(wakeword_model_location, alert.play)
    mic.start()
    while True:
        try:
            time.sleep(1)
            logger.info('foreground')
        except KeyboardInterrupt:
            break
    mic.stop()



from langchain.chat_models.base import BaseChatModel
from langchain_community.chat_models.ollama import ChatOllama
from langchain_openai import ChatOpenAI
from langchain_together import ChatTogether


class LLMFactory:
    def build(
        self,
        model_name: str,
        model_backend: str,
        context_size: int,
        temperature: float = 0.7,
        url: str = '',
        api_key: str = '',
    ) -> BaseChatModel:
        if model_backend == 'ollama':
            url = url if url else 'http://localhost:11434'
            llm = ChatOllama(
                model=model_name,
                num_ctx=context_size,
                temperature=temperature,
                base_url=url,
            )
        elif model_backend == 'openai':
            url = url if url else None
            llm = ChatOpenAI(
                model=model_name,
                max_tokens=context_size,
                temperature=temperature,
                openai_api_key=api_key,
                openai_api_base=url,
            )
        elif model_backend == 'together':
            url = url if url else 'https://api.together.xyz/inference'
            llm = ChatTogether(
                model=model_name,
                max_tokens=context_size,
                temperature=temperature,
                together_api_key=api_key,
                base_url=url,
            )
        elif model_backend == 'llamacpp':
            url = url if url else 'http://localhost:8000/v1'
            llm = ChatOpenAI(
                model=model_name,
                max_tokens=context_size,
                temperature=temperature,
                openai_api_key=api_key,
                openai_api_base=url,
            )
        else:
            raise ValueError('Unsuported model type')

        return llm



import audioop
from io import BytesIO
import time
import wave
from loguru import logger
import pyaudio


class Microphone:
    def __init__(self):
        self.chunk = 1024
        self.format = pyaudio.paInt16
        self.channels = 1
        self.rate = 44100
        self._pyaudio = pyaudio.PyAudio()

    def calibrate(self, interval_seconds: float) -> int:
        logger.info('[STARTED] calibrating microphone')

        stream = self._pyaudio.open(
            format=self.format,
            channels=self.channels,
            rate=self.rate,
            input=True,
            frames_per_buffer=self.chunk,
        )

        calibration_started_time = time.time()
        ambient_noise_data = []
        elapsed_calibration_time = 0

        while interval_seconds > elapsed_calibration_time:
            data = stream.read(self.chunk)
            rms = audioop.rms(data, 2)
            ambient_noise_data.append(rms)
            elapsed_calibration_time = time.time() - calibration_started_time

        stream.stop_stream()
        stream.close()

        silence_threshold = max(ambient_noise_data)

        logger.info('[FINISHED] calibrating microphone')

        return silence_threshold

    def record(
        self,
        max_duration=60,
        max_silence_seconds=2,
        patience=20,
        silence_threshold=-1,
    ) -> BytesIO:
        logger.info('[STARTED] Microfone recording')

        stream = self._pyaudio.open(
            format=self.format,
            channels=self.channels,
            rate=self.rate,
            input=True,
            frames_per_buffer=self.chunk,
        )

        frames = []
        record_time = 0
        last_speach = time.time()
        consecutive_silece_windows = 0

        while record_time < max_duration:
            data = stream.read(self.chunk)
            frames.append(data)

            rms = audioop.rms(data, 2)
            last_speach_time_delta = time.time() - last_speach

            if rms > silence_threshold:
                state = 'listening'
                last_speach = time.time()
                consecutive_silece_windows = 0
            else:
                state = 'silence'
                consecutive_silece_windows += 1

                if consecutive_silece_windows > patience:
                    if last_speach_time_delta > max_silence_seconds:
                        break
            record_time += len(data) / self.rate

        stream.stop_stream()
        stream.close()

        sample_size = self._pyaudio.get_sample_size(self.format)
        bytesio = BytesIO()
        wf = wave.open(bytesio, 'wb')
        wf.setnchannels(self.channels)
        wf.setsampwidth(sample_size)
        wf.setframerate(self.rate)
        wf.writeframes(b''.join(frames))
        wf.close()
        bytesio.seek(0)

        logger.info('[ENDED] Microfone recording')

        return bytesio



from typing import Sequence

from .coordinate import Coordinate
from .movement import Movement


class SequenceMovement(Movement):
    def __init__(self, movements: Sequence[Movement]) -> None:
        super().__init__()
        self._movements = movements
        self._movement_index = 0

    def step(self) -> Coordinate:
        current_movement = self._movements[self._movement_index]
        current_position = current_movement.step()

        if current_movement.is_at_endding():
            self._movement_index += 1

        self._movement_index = min(
            self._movement_index, len(self._movements) - 1
        )

        return current_position

    def step_reverse(self) -> Coordinate:
        current_movement = self._movements[self._movement_index]
        current_position = current_movement.step_reverse()

        if current_movement.is_at_beginning():
            self._movement_index -= 1

        self._movement_index = max(0, self._movement_index)

        return current_position

    def is_at_beginning(self) -> bool:
        current_movement = self._movements[self._movement_index]
        is_first_index = self._movement_index == 0
        return is_first_index and current_movement.is_at_beginning()

    def is_at_endding(self) -> bool:
        current_movement = self._movements[self._movement_index]
        is_last_index = self._movement_index >= len(self._movements) - 1
        return is_last_index and current_movement.is_at_endding()



from concurrent.futures import Future, ThreadPoolExecutor
from pathlib import Path
from typing import Callable

from sprite_ai.language.chat_message import ChatMessage
from sprite_ai.language.chat_state import ChatState
from sprite_ai.language.language_model import LanguageModel


class ChatSession:
    def __init__(
        self,
        chat_state: ChatState | None = None,
    ) -> None:
        if chat_state is None:
            chat_state = ChatState()
        self.chat_state = chat_state

    def add_message(self, message: ChatMessage):
        self.chat_state.chat_history.append(message)

    @property
    def messages(self):
        return self.chat_state.chat_history

    def save_state(self, state_location: str | Path):
        self.chat_state.to_file(state_location)

    def load_state(self, state_location: str | Path):
        self.chat_state = ChatState.from_file(state_location)

    def clear_state(self):
        self.chat_state = ChatState()



from pathlib import Path
import sys
import threading
from typing import Any, Callable

import numpy as np
from pydantic import BaseModel
from PyQt5 import QtWidgets
from PyQt5.QtGui import QIcon, QPixmap

from sprite_ai.movement import Movement
from sprite_ai.movement.coordinate import Coordinate
from sprite_ai.movement.linear_movement import LinearMovement
from sprite_ai.sprite_sheet.animation import Animation, AnimationController
from sprite_ai.sprite_sheet.sprite_sheet import SpriteSheetMetadata
from sprite_ai.sprite_widget import SpriteWidgetQt


class SpriteGui:
    def __init__(
        self,
        screen_size: tuple[int, int],
        sprite_sheet_metadata: SpriteSheetMetadata,
        animations: dict[str, Animation],
        on_position_updated: Callable | None = None,
        on_clicked: Callable | None = None,
        icon_location: str | Path = '',
    ):
        self.icon_location = icon_location
        self.screen_size = screen_size
        self.sprite_widget = SpriteWidgetQt()
        sprite_sheet_image = QPixmap(sprite_sheet_metadata.path)
        self.animation_controller = AnimationController(
            sprite_sheet_image, sprite_sheet_metadata, animations
        )
        self.image_update_rate: float | int = 0.1
        self.position_update_rate: float | int = 0.2
        self._image_update_timer: threading.Timer | None = None
        self._position_update_timer: threading.Timer | None = None
        self._movement: LinearMovement | None = None
        self.on_position_updated = on_position_updated
        if on_clicked is not None:
            self.sprite_widget.qwidget.mouseReleaseEvent = on_clicked

    def _update_image_loop(self):
        self._image_update_timer = threading.Timer(
            self.image_update_rate, self._update_image_loop
        )
        self._image_update_timer.start()
        self.sprite_widget.image = self.animation_controller.frame

    def _update_position_loop(self):
        self._position_update_timer = threading.Timer(
            self.position_update_rate, self._update_position_loop
        )
        self._position_update_timer.start()
        if self._movement is None:
            return

        x, y = self._movement.step()
        new_position = Coordinate(int(x), int(y))
        current_position = self.sprite_widget.position
        # has_position_changed = current_position != new_position
        self.sprite_widget.position = new_position

        if self.on_position_updated != None:
            position_update_message = {
                'old_position': current_position,
                'new_position': new_position,
            }
            self.on_position_updated(position_update_message)

    def gui_loop(self):
        width, height = self.screen_size
        self.sprite_widget.position = (width // 2, height)
        self.sprite_widget.show()
        self._update_image_loop()
        self._update_position_loop()
        self.animation_controller.play()

    def set_movement(self, movement: LinearMovement):
        self.animation_controller.set_orientation(movement.orientation)
        self._movement = movement

    def set_animation(self, name: str):
        self.animation_controller.set_animation(name)

        if self._movement:
            orientation = self._movement.orientation
            self.animation_controller.set_orientation(orientation)

    def run(self):
        self.gui_loop()

    def shutdown(self):
        self._image_update_timer.cancel()
        self._image_update_timer.join()
        self._position_update_timer.cancel()
        self._position_update_timer.join()



from collections import defaultdict
from threading import Lock
from typing import Any, Callable, DefaultDict

from loguru import logger

Callback = Callable[[Any], None]


class EventManager:
    def __init__(self, skip_topic_log: set[str] | None = None) -> None:
        self.skip_topic_log: set[str] = (
            skip_topic_log if skip_topic_log else set()
        )
        self.inscriptions: DefaultDict[str, list[Callback]] = defaultdict(list)
        self._lock = Lock()

    def publish(self, topic: str, message: Any = None) -> None:
        """Publish message to a topic

        Args:
            message (Any): A object to be passed as message
            topic (str): A topic to publish the object to
        """
        if topic not in self.skip_topic_log:
            logger.debug(f'[{topic}] {message}')

        with self._lock:
            topic_callbacks = self.inscriptions[topic]

        for callback in topic_callbacks:
            callback(message)

    def subscribe(self, topic: str, callback: Callback) -> None:
        """Subscribe a callable to a topic

        Args:
            callback (Callback): a callable to be called when something is posted to the topic
            topic (str): the topic to link the callback to
        """
        with self._lock:
            self.inscriptions[topic].append(callback)

    def subscribers(self, topic: str) -> list:
        """Retrive subscribers of a given topic

        Args:
            topic (str): topic name

        Returns:
            list: callbacks subscribed to topic
        """
        return self.inscriptions[topic]



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/chat_window.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(800, 600)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.lv_chat_history = QtWidgets.QListView(self.centralwidget)
        self.lv_chat_history.setMinimumSize(QtCore.QSize(0, 10))
        self.lv_chat_history.setObjectName("lv_chat_history")
        self.verticalLayout.addWidget(self.lv_chat_history)
        self.hl_user_interface = QtWidgets.QHBoxLayout()
        self.hl_user_interface.setContentsMargins(-1, 0, -1, -1)
        self.hl_user_interface.setObjectName("hl_user_interface")
        self.te_chatinput = QtWidgets.QTextEdit(self.centralwidget)
        self.te_chatinput.setMaximumSize(QtCore.QSize(16777215, 80))
        self.te_chatinput.setObjectName("te_chatinput")
        self.hl_user_interface.addWidget(self.te_chatinput)
        self.pb_send = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pb_send.sizePolicy().hasHeightForWidth())
        self.pb_send.setSizePolicy(sizePolicy)
        self.pb_send.setObjectName("pb_send")
        self.hl_user_interface.addWidget(self.pb_send)
        self.verticalLayout.addLayout(self.hl_user_interface)
        MainWindow.setCentralWidget(self.centralwidget)
        self.mb_menubar = QtWidgets.QMenuBar(MainWindow)
        self.mb_menubar.setGeometry(QtCore.QRect(0, 0, 800, 22))
        self.mb_menubar.setObjectName("mb_menubar")
        self.m_menu = QtWidgets.QMenu(self.mb_menubar)
        self.m_menu.setObjectName("m_menu")
        MainWindow.setMenuBar(self.mb_menubar)
        self.sb_statusbar = QtWidgets.QStatusBar(MainWindow)
        self.sb_statusbar.setObjectName("sb_statusbar")
        MainWindow.setStatusBar(self.sb_statusbar)
        self.a_exit = QtWidgets.QAction(MainWindow)
        self.a_exit.setObjectName("a_exit")
        self.actionHide_window = QtWidgets.QAction(MainWindow)
        self.actionHide_window.setObjectName("actionHide_window")
        self.a_new_chat = QtWidgets.QAction(MainWindow)
        self.a_new_chat.setObjectName("a_new_chat")
        self.a_clear_chat = QtWidgets.QAction(MainWindow)
        self.a_clear_chat.setObjectName("a_clear_chat")
        self.a_settings = QtWidgets.QAction(MainWindow)
        self.a_settings.setObjectName("a_settings")
        self.m_menu.addAction(self.a_settings)
        self.m_menu.addAction(self.a_clear_chat)
        self.m_menu.addAction(self.a_exit)
        self.mb_menubar.addAction(self.m_menu.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Sprite AI"))
        self.pb_send.setText(_translate("MainWindow", "Send"))
        self.m_menu.setTitle(_translate("MainWindow", "Menu"))
        self.a_exit.setText(_translate("MainWindow", "Exit"))
        self.a_exit.setShortcut(_translate("MainWindow", "Ctrl+Q"))
        self.actionHide_window.setText(_translate("MainWindow", "Hide window"))
        self.a_new_chat.setText(_translate("MainWindow", "New chat"))
        self.a_clear_chat.setText(_translate("MainWindow", "Clear chat"))
        self.a_settings.setText(_translate("MainWindow", "Settings"))



from pathlib import Path
import soundfile as sf
import sounddevice as sd


class Sound:
    def __init__(self, sound_location: str | Path):
        self.sound_location = sound_location
        self.audio_data, self.samplerate = sf.read(
            sound_location, always_2d=True
        )

    def play(self, blocking: bool = False):
        sd.play(self.audio_data, samplerate=self.samplerate, blocking=blocking)



from __future__ import annotations
from multiprocessing import Process
import multiprocessing
import time


import uvicorn

from llama_cpp.server.app import create_app
from llama_cpp.server.settings import Settings
from huggingface_hub import hf_hub_download


class LanguageModelServer:
    def __init__(
        self,
        hf_repo_id: str,
        file_name: str,
        host: str = 'localhost',
        port: int = 8000,
        context_size: int = 8192,
        cache: bool = True,
        model_alias: str | None = None,
        verbose: bool = False,
    ) -> None:
        self._server_process = Process(
            name='language_model_server',
            target=self._run,
            args=(
                hf_repo_id,
                file_name,
                host,
                port,
                context_size,
                cache,
                model_alias,
                verbose,
            ),
        )

    def start(self):
        self._server_process.start()

    def stop(self):
        self._server_process.terminate()
        self._server_process.join()
        self._server_process.close()

    @staticmethod
    def _run(
        hf_repo_id: str,
        file_name: str,
        host: str,
        port: int,
        context_size: int,
        cache: bool,
        model_alias: str | None,
        verbose: bool,
    ):
        model_location = hf_hub_download(
            repo_id=hf_repo_id,
            filename=file_name,
        )
        # gets the number of physical cpu cores
        cpu_count = multiprocessing.cpu_count()
        settings = Settings(
            model=model_location,
            model_alias=model_alias,
            n_ctx=context_size,
            host=host,
            port=port,
            cache=cache,
            verbose=verbose,
            n_threads=cpu_count,
        )
        app = create_app(
            settings=settings,
        )
        uvicorn.run(app)


if __name__ == '__main__':
    server = LanguageModelServer(
        'cognitivecomputations/dolphin-2.9-llama3-8b-gguf',
        'dolphin-2.9-llama3-8b-q4_K_M.gguf',
    )
    server.start()
    time.sleep(10)
    server.stop()



from sprite_ai.sprite_sheet.animation import Animation

ANIMATIONS = {
    'idle': Animation(0, 0, 0.2),
    'walking': Animation(0, 3, 0.2),
    'jumping': Animation(2, 4, 0.2),
    'playing': Animation(30, 33, 0.2),
    'sliding': Animation(19, 21, 0.2),
    'thinking': Animation(14, 17, 0.2),
    'appearing': Animation(26, 29, 1),
    'playing': Animation(29, 33, 0.2),
    'walking_upright': Animation(33, 36, 0.2),
    'laying_down': Animation(17, 17, 0.2),
}
